unit ADODBAccess;

interface

uses Windows, Classes, db, ADOdb, SysUtils;

resourcestring
  SAutoSessionExclusive = 'Cannot enable AutoSessionName property with more than one session on a form or data-module';
  SAutoSessionExists = 'Cannot add a session to the form or data-module while session ''%s'' has AutoSessionName enabled';
  SAutoSessionActive = 'Cannot modify SessionName while AutoSessionName is enabled';
  SDuplicateDatabaseName = 'Duplicate database name ''%s''';
  SDuplicateSessionName = 'Duplicate session name ''%s''';
  SInvalidSessionName = 'Invalid session name %s';
  SDatabaseNameMissing = 'Database name missing';
  SSessionNameMissing = 'Session name missing';
  SDatabaseOpen = 'Cannot perform this operation on an open database';
  SDatabaseClosed = 'Cannot perform this operation on a closed database';
  SDatabaseHandleSet = 'Database handle owned by a different session';
  SSessionActive = 'Cannot perform this operation on an active session';
  SHandleError = 'Error creating cursor handle';
  SInvalidFloatField = 'Cannot convert field ''%s'' to a floating point value';
  SInvalidIntegerField = 'Cannot convert field ''%s'' to an integer value';
  STableMismatch = 'Source and destination tables are incompatible';
  SFieldAssignError = 'Fields ''%s'' and ''%s'' are not assignment compatible';
  SNoReferenceTableName = 'ReferenceTableName not specified for field ''%s''';
  SCompositeIndexError = 'Cannot use array of Field values with Expression Indices';
  SInvalidBatchMove = 'Invalid batch move parameters';
  SEmptySQLStatement = 'No SQL statement available';
  SNoParameterValue = 'No value for parameter ''%s''';
  SNoParameterType = 'No parameter type for parameter ''%s''';
  SLoginError = 'Cannot connect to database ''%s''';
  SInitError = 'An error occurred while attempting to initialize the Borland Database Engine (error $%.4x)';
  SDatabaseEditor = 'Da&tabase Editor...';
  SExplore = 'E&xplore';
  SLinkDetail = '''%s'' cannot be opened';
  SLinkMasterSource = 'The MasterSource property of ''%s'' must be linked to a DataSource';
  SLinkMaster = 'Unable to open the MasterSource Table';
  SGQEVerb = 'S&QL Builder...';
  SBindVerb = 'Define &Parameters...';
  SIDAPILangID = '0009';
  SDisconnectDatabase = 'Database is currently connected. Disconnect and continue?';
  SBDEError = 'BDE error $%.4x';
  SLookupSourceError = 'Unable to use duplicate DataSource and LookupSource';
  SLookupTableError = 'LookupSource must be connected to TTable component';
  SLookupIndexError = '%s must be the lookup table''s active index';
  SParameterTypes = ';Input;Output;Input/Output;Result';
  SInvalidParamFieldType = 'Must have a valid field type selected';
  STruncationError = 'Parameter ''%s'' truncated on output';
  SDataTypes = ';String;SmallInt;Integer;Word;Boolean;Float;Currency;BCD;Date;Time;DateTime;;;;Blob;Memo;Graphic;;;;;Cursor;';
  SResultName = 'Result';
  SDBCaption = '%s%s%s Database';
  SParamEditor = '%s%s%s Parameters';
  SIndexFilesEditor = '%s%s%s Index Files';
  SNoIndexFiles = '(None)';
  SIndexDoesNotExist = 'Index does not exist. Index: %s';
  SNoTableName = 'Missing TableName property';
  SNoDataSetField = 'Missing DataSetField property';
  SBatchExecute = 'E&xecute';
  SNoCachedUpdates = 'Not in cached update mode';
  SInvalidAliasName = 'Invalid alias name %s';
  SNoFieldAccess = 'Cannot access field ''%s'' in a filter';
  SUpdateSQLEditor = '&UpdateSQL Editor...';
  SNoDataSet = 'No dataset association';
  SUntitled = 'Untitled Application';
  SUpdateWrongDB = 'Cannot update, %s is not owned by %s';
  SUpdateFailed = 'Update failed';
  SSQLGenSelect = 'Must select at least one key field and one update field';
  SSQLNotGenerated = 'Update SQL statements not generated, exit anyway?';
  SSQLDataSetOpen = 'Unable to determine field names for %s';
  SLocalTransDirty = 'The transaction isolation level must be dirty read for local databases';
  SMissingDataSet = 'Missing DataSet property';
  SNoProvider = 'No provider available';
  SNotAQuery = 'Dataset is not a query';

type
  TADOUpdateQuery = class;
  TADOUpdateSQL = class;

  TADOUpdateQuery = class (TADOQuery)  // For proper operation (logic) is not desirable to change the required fields
                                       // 󥿽Tާ@]޿^Oiܩһݪrq
  private
    DelRecords: TADOQuery;
    FUpdateObject: TADOUpdateSQL;
    procedure SetUpdateObject(Value: TADOUpdateSQL);
    procedure ClearBuffer;  // physically remove deleted records from the buffer
    procedure InitBuffer;   // create a dataset that pomeschayuisya deleted records
    procedure FillBuffer;   // transfer buffering remote
    procedure ApplyDelUpdates;
  protected
    procedure InternalDelete; override;
  public
    constructor Create (AOwner: TComponent); override;
    destructor Destroy; override;
    procedure ApplyUpdates;  // After a successful buffer deleted records will be empty and needs CommitUpdates
                             // As the status of "untouched" records are not changed
                             // (for example inserting a record: it will be as much time
                             // Was called ApplyUpdates. That is wrong :-(, people know about it
    procedure CancelUpdates; // resets internal flags ado (inserted, modified) and remote reset
    procedure CommitUpdates; // resets internal flags ado (inserted, modified)
  published
    property UpdateObject: TADOUpdateSQL read FUpdateObject write SetUpdateObject;
  end;


  TADOUpdateSQL = class(TComponent)
  private
    FDataSet: TADOUpdateQuery;
    FQueries: array[TUpdateKind] of TADOQuery;
    FSQLText: array[TUpdateKind] of TStrings;
    function GetQuery(UpdateKind: TUpdateKind): TADOQuery;
    function GetSQLIndex(Index: Integer): TStrings;
    procedure SetSQL(UpdateKind: TUpdateKind; Value: TStrings);
    procedure SetSQLIndex(Index: Integer; Value: TStrings);
  protected
    function GetSQL(UpdateKind: TUpdateKind): TStrings; virtual;
    function GetDataSet: TADOUpdateQuery; virtual;
    procedure SetDataSet(ADataSet: TADOUpdateQuery); virtual;
    procedure SQLChanged(Sender: TObject);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Apply(UpdateKind: TUpdateKind);  virtual; // ܪAO
    procedure ExecSQL(UpdateKind: TUpdateKind);         // եsetparamsAܪAO
    procedure SetParams(UpdateKind: TUpdateKind);       // RѼ
    property DataSet: TADOUpdateQuery read GetDataSet write SetDataSet;
    property Query[UpdateKind: TUpdateKind]: TADOQuery read GetQuery;
    property SQL[UpdateKind: TUpdateKind]: TStrings read GetSQL write SetSQL;
  published
    property ModifySQL: TStrings index 0 read GetSQLIndex write SetSQLIndex;
    property InsertSQL: TStrings index 1 read GetSQLIndex write SetSQLIndex;
    property DeleteSQL: TStrings index 2 read GetSQLIndex write SetSQLIndex;
  end;


procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('dbGO', [TADOUpdateQuery, TADOUpdateSQL ]);
end;

{ TADOUpdateSQL }


constructor TADOUpdateSQL.Create(AOwner: TComponent);
var
  UpdateKind: TUpdateKind;
begin
  inherited Create(AOwner);
  for UpdateKind := Low(TUpdateKind) to High(TUpdateKind) do
  begin
    FSQLText[UpdateKind] := TStringList.Create;
    TStringList(FSQLText[UpdateKind]).OnChange := SQLChanged;
  end;
end;

destructor TADOUpdateSQL.Destroy;
var
  UpdateKind: TUpdateKind;
begin
  if Assigned(FDataSet) and (FDataSet.UpdateObject = Self) then
    FDataSet.UpdateObject := nil;
  for UpdateKind := Low(TUpdateKind) to High(TUpdateKind) do
    FSQLText[UpdateKind].Free;
  inherited Destroy;
end;

procedure TADOUpdateSQL.ExecSQL(UpdateKind: TUpdateKind);
begin
  with Query[UpdateKind] do
  begin
    ExecSQL;
    if RowsAffected <> 1 then DatabaseError(SUpdateFailed);
  end;
end;

function TADOUpdateSQL.GetQuery(UpdateKind: TUpdateKind): TADOQuery;//TADOUpdateQuery;
begin
  if not Assigned(FQueries[UpdateKind]) then
  begin
    FQueries[UpdateKind] := TADOQuery.Create(Self);    // Create TADOQuery
    FQueries[UpdateKind].Connection := FDataSet.Connection; // attached to the main Connect
    FQueries[UpdateKind].SQL.Assign(FSQLText[UpdateKind]);  // passing in the appropriate request
  end;
  Result := FQueries[UpdateKind];
end;

function TADOUpdateSQL.GetSQL(UpdateKind: TUpdateKind): TStrings;
begin
  Result := FSQLText[UpdateKind];
end;

function TADOUpdateSQL.GetSQLIndex(Index: Integer): TStrings;
begin
  Result := FSQLText[TUpdateKind(Index)];
end;

function TADOUpdateSQL.GetDataSet: TADOUpdateQuery;
begin
  Result := FDataSet;
end;

procedure TADOUpdateSQL.SetDataSet(ADataSet: TADOUpdateQuery);
begin
  FDataSet := ADataSet;
end;

procedure TADOUpdateSQL.SetSQL(UpdateKind: TUpdateKind; Value: TStrings);
begin
  FSQLText[UpdateKind].Assign(Value);
end;

procedure TADOUpdateSQL.SetSQLIndex(Index: Integer; Value: TStrings);
begin
  SetSQL(TUpdateKind(Index), Value);
end;

procedure TADOUpdateSQL.SQLChanged(Sender: TObject);
var
  UpdateKind: TUpdateKind;
begin
  for UpdateKind := Low(TUpdateKind) to High(TUpdateKind) do
    if Sender = FSQLText[UpdateKind] then
    begin
      if Assigned(FQueries[UpdateKind]) then
      begin
        FQueries[UpdateKind].Parameters.Clear;
        FQueries[UpdateKind].SQL.Assign(FSQLText[UpdateKind]);
      end;
      Break;
    end;
end;

procedure TADOUpdateSQL.SetParams(UpdateKind: TUpdateKind);
var
  i: Integer;
  Old: Boolean;
  Param: TParameter;
  PName: string;
  Field: TField;
  tQuery: TADOQuery;
begin
  if not Assigned(FDataSet) then Exit;
  tQuery:=Query[UpdateKind];
    for i := 0 to tQuery.Parameters.Count - 1 do
    begin
      Param := tQuery.Parameters.Items[I];
      PName := Param.Name;
      Old := (CompareText(Copy(PName, 1, 4), 'OLD_') = 0);
      if Old then System.Delete(PName, 1, 4);
      Old:= Old and (UpdateKind<>ukDelete);  // Old flag reset for remote (although 'Old_' is, for compatibility)
      Field := FDataSet.FindField(PName);
      if not Assigned(Field) then Continue;
      if Old then // if the parameter is filled with the old value
          begin
          Param.DataType:=Field.DataType;
          Param.Value:=Field.OldValue;
          end
      else
            begin
            Param.DataType:=Field.DataType;
            Param.Value:=Field.Value;
            end;
    end;
end;

procedure TADOUpdateSQL.Apply(UpdateKind: TUpdateKind);
begin
  SetParams(UpdateKind);
  ExecSQL(UpdateKind);
end;


{ TADOUpdateQuery }

procedure TADOUpdateQuery.ApplyUpdates;
var
  BM: TBookmark;
  tFiltered: boolean;
begin // This code will need to finalize
DisableControls;
BM:=GetBookmark;
tFiltered:=Filtered;
Filtered:=False;  // filter to turn it off, that is, the change will be subject to the "untouched" record
try
  try 
    First; // move the cursor to the last to the first machine made Post
    while not Eof do
      begin
      if RecordStatus = [rsNew] then
        UpdateObject.Apply(ukInsert) else
           if RecordStatus = [rsModified] then
              UpdateObject.Apply(ukModify);
      Next;
      end;
      if not DelRecords.IsEmpty then
      begin
      ApplyDelUpdates;
      ClearBuffer; // if successful then remove from the buffer of deleted records
      end;
  finally
      Filtered:=tFiltered;
      GotoBookmark(BM);
      FreeBookmark(BM);
      EnableControls;
  end;
except
  on E: EDatabaseError do raise EDatabaseError.Create(E.Message);
end;
end;

procedure TADOUpdateQuery.CancelUpdates;
begin
  inherited CancelUpdates;
  ClearBuffer; // kill all deleted records of the corresponding table
end;

constructor TADOUpdateQuery.Create(AOwner: TComponent);
begin
 inherited Create(AOwner);
 LockType:=ltBatchOptimistic;
 DelRecords:=TADOQuery.Create(self);
end;

procedure TADOUpdateQuery.InitBuffer;
begin
if DelRecords.FieldDefs.Count = 0 then
  begin
  DelRecords.Connection :=Connection;
  DelRecords.FieldDefs.Assign(FieldDefs);
  TADODataSet(DelRecords).CreateDataSet;
  end;
end;

destructor TADOUpdateQuery.Destroy;
begin
  DelRecords.Free;
  inherited Destroy;
end;


procedure TADOUpdateQuery.InternalDelete;
begin
 if not (RecordStatus = [rsNew]) then
   FillBuffer;
 inherited InternalDelete;   // Removed
end;


procedure TADOUpdateQuery.SetUpdateObject(Value: TADOUpdateSQL);
begin
  if Value <> FUpdateObject then
  begin
    if Assigned(FUpdateObject) and (FUpdateObject.DataSet = Self) then
      FUpdateObject.DataSet := nil;
    FUpdateObject := Value;
    if Assigned(FUpdateObject) then
    begin
      {If other dataset already refers to this update, we delete the link}
      if Assigned(FUpdateObject.DataSet) and
        (FUpdateObject.DataSet <> Self) then
        FUpdateObject.DataSet.UpdateObject:= nil;
      FUpdateObject.DataSet := Self;
    end;
  end;
end;

procedure TADOUpdateQuery.ClearBuffer;
begin
  if not DelRecords.IsEmpty then
   while not DelRecords.Bof do   // cursor is always on the last record (added to buffer)
   DelRecords.Delete;            // remove entries
end;

procedure TADOUpdateQuery.FillBuffer;
var
  i: integer;
begin
  InitBuffer; // initialize buffer
  DelRecords.Append;
    for i:= 0 to Fields.Count-1 do
        if Fields[i].FieldKind = fkData then      // Only real data (bad?)
         DelRecords.Fields[i].Assign(Fields[i]);  // Moved
  DelRecords.Post;
end;

procedure TADOUpdateQuery.ApplyDelUpdates;
var
  tQuery: TADOUpdateQuery;
begin
  tQuery:=FUpdateObject.FDataSet; // Query the main green triangle icon
  FUpdateObject.FDataSet:=TADOUpdateQuery(DelRecords); // substitute deleted(FDataSet.DelRecords) TADOQuery
  DelRecords.First;
  while not DelRecords.Eof do
    begin
    UpdateObject.Apply(ukDelete);
    DelRecords.Next;
    end;
  FUpdateObject.FDataSet:=tQuery; // Gets the Query
end;

procedure TADOUpdateQuery.CommitUpdates;
begin
Requery;
end;

end.
